pub struct Bvh2ValidationResult {
pub direct_layout: bool,
pub require_tight_fit: bool,
pub discovered_primitives: HashSet<u32>,
pub discovered_nodes: HashSet<u32>,
pub node_count: usize,
pub leaf_count: usize,
pub prim_count: usize,
pub max_depth: u32,
pub nodes_at_depth: HashMap<u32, u32>,
pub leaves_at_depth: HashMap<u32, u32>,
}Expand description
Result of Bvh2 validation. Contains various bvh stats.
Fields
direct_layout: boolThe primitives are already laid out in bvh.primitive_indices order.
require_tight_fit: boolRequire validation to ensure aabbs tightly fit children and primitives.
discovered_primitives: HashSet<u32>Set of primitives discovered though validation traversal.
discovered_nodes: HashSet<u32>Set of nodes discovered though validation traversal.
node_count: usizeTotal number of nodes discovered though validation traversal.
leaf_count: usizeTotal number of leaves discovered though validation traversal.
prim_count: usizeTotal number of primitives discovered though validation traversal.
max_depth: u32Maximum hierarchical BVH depth discovered though validation traversal.
nodes_at_depth: HashMap<u32, u32>Quantity of nodes found at each depth though validation traversal.
leaves_at_depth: HashMap<u32, u32>Quantity of leaves found at each depth though validation traversal.